import textwrap

PROCEDURE_TEMPLATE = textwrap.dedent(
    """\
    CREATE OR REPLACE FUNCTION {audit_function_name}() RETURNS TRIGGER AS ${trigger_name}$
    BEGIN
        {check_settings}

        IF (TG_OP = 'DELETE') THEN
            INSERT INTO {audit_table_full_name} ({audit_columns}) SELECT {deletion_elements};
        ELSIF (TG_OP = 'UPDATE') THEN
            INSERT INTO {audit_table_full_name} ({audit_columns}) SELECT {updation_elements};
        ELSIF (TG_OP = 'INSERT') THEN
            INSERT INTO {audit_table_full_name} ({audit_columns}) SELECT {insertion_elements};
        END IF;
        RETURN NULL; -- result is ignored since this is an AFTER trigger
    END;
    ${trigger_name}$ LANGUAGE plpgsql;

    DROP TRIGGER IF EXISTS {trigger_name} ON {table_full_name};

    CREATE TRIGGER {trigger_name}
    AFTER INSERT OR UPDATE OR DELETE ON {table_full_name}
    FOR EACH ROW EXECUTE PROCEDURE {audit_function_name}();
    """
)


DROP_PROCEDURE_TEMPLATE = textwrap.dedent(
    """\
    DROP TRIGGER IF EXISTS {trigger_name} ON {table_full_name};
    DROP FUNCTION IF EXISTS {function_name};
    """
)


def make_audit_procedure(
    procedure_name,
    trigger_name,
    check_settings,
    audit_table_full_name,
    table_full_name,
    audit_columns,
    deletion_elements,
    updation_elements,
    insertion_elements,
):
    """Return the string

    :param procedure_name: The name for the procedure
    :param trigger_name:  The name for the trigger
    :param check_settings: A list of settings checks.
        Used to validate that the settings are non-empty.
    :param audit_table_full_name: The full name (including schema) of the audit table
    :param table_full_name: The full name (including schema) of the audited table
    :param audit_columns: A list of columns to be audited
    :param deletion_elements: A list of column expressions
        suitable to use in the VALUES block of an insert statement.
    :param updation_elements: A list of column expressions
        suitable to use in the VALUES block of an insert statement.
    :param insertion_elements: A list of column expressions
        suitable to use in the VALUES block of an insert statement.
    :return: A :class:`str` of the full DDL needed to be executed
        to create the procedure and trigger.
    """

    deletion_elements_str = ", ".join(deletion_elements)
    updation_elements_str = ", ".join(updation_elements)
    insertion_elements_str = ", ".join(insertion_elements)
    columns_str = ", ".join(audit_columns)
    check_settings_str = "\n    ".join(check_settings)

    return PROCEDURE_TEMPLATE.format(
        audit_function_name=procedure_name,
        trigger_name=trigger_name,
        check_settings=check_settings_str,
        audit_table_full_name=audit_table_full_name,
        table_full_name=table_full_name,
        audit_columns=columns_str,
        deletion_elements=deletion_elements_str,
        updation_elements=updation_elements_str,
        insertion_elements=insertion_elements_str,
    )


def make_drop_audit_procedure(function_name, trigger_name, table_full_name):
    return DROP_PROCEDURE_TEMPLATE.format(
        function_name=function_name,
        trigger_name=trigger_name,
        table_full_name=table_full_name,
    )



__version__ = "0.5.2"

__all__ = [
    "set_session_vars",
    "enable",
    "install_audit_triggers",
    "uninstall_audit_triggers",
]

from .session import set_session_vars
from .plugin import enable
from .ddl import install_audit_triggers, uninstall_audit_triggers



from alembic.operations.ops import ExecuteSQLOp


class ReversableExecute(ExecuteSQLOp):
    def __init__(self, sqltext, reverse_ddl, execution_options=None):
        super(ReversableExecute, self).__init__(sqltext, execution_options)
        self._reverse_ddl = reverse_ddl
        self._execution_options = execution_options

    def reverse(self):
        return ExecuteSQLOp(self._reverse_ddl, self._execution_options)


def compare_for_table(
    autogen_context, modify_table_ops, schema, tname, conn_table, metadata_table
):
    # Early exit if there is no metadata table or we can't get it's info.
    # This handles alembic version tables, but also others.
    if metadata_table is None or not hasattr(metadata_table, "info"):
        return

    if metadata_table.info.get("audit.is_audit_table"):
        # Case when the audit table is new
        if conn_table is None:
            modify_table_ops.ops.append(
                ReversableExecute(
                    sqltext=metadata_table.info["audit.target_table"].info[
                        "audit.create_ddl"
                    ],
                    reverse_ddl=metadata_table.info["audit.target_table"].info[
                        "audit.drop_ddl"
                    ],  # TODO: Use reflected DDL from the database
                )
            )

        # Case when the audit table exists already
        else:
            existing_audit_table_columns = set(
                [col.name for col in conn_table.columns.values()]
            )
            defined_audit_table_columns = set(
                [col.name for col in metadata_table.columns.values()]
            )

            # Case when the audit table has new columns.
            # TODO: Check if the settings types have changed.
            #  Not 100% sure how this would be done.
            if existing_audit_table_columns.symmetric_difference(
                defined_audit_table_columns
            ):
                modify_table_ops.ops.append(
                    ReversableExecute(
                        sqltext=metadata_table.info["audit.target_table"].info[
                            "audit.create_ddl"
                        ],
                        reverse_ddl=metadata_table.info["audit.target_table"].info[
                            "audit.drop_ddl"
                        ],  # TODO: Use reflected DDL from the database
                    )
                )

            # Case when the audit table triggers _should_ already be installed
            else:
                # TODO: Check to ensure that the triggers are installed
                pass

    elif not metadata_table.info.get("audit.is_audited"):
        # TODO: Drop the triggers if they exist.
        pass



def set_session_var_stmt(**kwargs):
    """Returns proper sql statements for setting session settings.

    Namespaces all settings under `audit.*` namespace.

    e.g.

        .. code-block:: python

            set_session_var_stmt(foo='bar', baz='foobaz')
            # set local "audit.foo" = 'bar'; set local "audit.baz" = 'foobaz';

    :param kwargs: key/value pairs of values to set.
    :return: a :class:`str`, valid to set the relevant settings.
    """
    format_str = "set local {} = {}"
    stmts = []
    for key, value in kwargs.items():
        stmts.append(format_str.format('"audit.{}"'.format(key), "'{}'".format(value)))
    return "; ".join(stmts) + ";"


def set_session_vars(connectable, **kwargs):
    """Wrapper to set session settings.

    This must be set *in* a transaction in order for these settings to be present.

    Typical use cases would be a username coming from a web request, or a request UUID
    or a script name.

    :param connectable: A connectable that we can execute on.
    :param kwargs: key/value pairs of values to set.
    :return: None
    """
    connectable.execute(set_session_var_stmt(**kwargs))



from sqlalchemy.engine import CreateEnginePlugin
from sqlalchemy.dialects.postgresql.base import PGDialect
from sqlalchemy_postgresql_audit.dialect import PGAdvancedInspector
from sqlalchemy_postgresql_audit.event_listeners import enable_event_listeners


def enable():
    """Enable the advanced inspector and enables sqlalchemy and alembic event listeners."""
    PGDialect.inspector = PGAdvancedInspector
    enable_event_listeners()


class AuditPlugin(CreateEnginePlugin):
    def __init__(self, url, kwargs):
        super(AuditPlugin, self).__init__(url, kwargs)
        enable()



from sqlalchemy.dialects.postgresql.base import PGInspector

DEFAULT_AUDIT_TABLE_NAMING_CONVENTION = "%(table_name)s_audit"
"""The audit table naming convetion. Change this at naming_conventions `audit.table` key."""

DEFAULT_AUDIT_TABLE_FUNCTION_NAMING_CONVENTION = "%(schema)s_%(table_name)s_audit"
"""The audit table naming convetion. Change this at naming_conventions `audit.table` key."""

DEFAULT_AUDIT_TABLE_TRIGGER_CONVENTION = "%(schema)s_%(table_name)s_audit"
"""The audit table naming convetion. Change this at naming_conventions `audit.table` key."""


class PGAdvancedInspector(PGInspector):
    """A subclass of :class:`sqlalchemy.dialects.postgresql.base.PGInspector`.

    Enables advanced database reflection.
    """

    def reflecttable(self, table, include_columns, *args, **kwargs):
        super(PGAdvancedInspector, self).reflecttable(
            table, include_columns, *args, **kwargs
        )
        # TODO: Retrieve trigger/procedure information for the particular table.



import threading

from sqlalchemy import Table
from sqlalchemy.events import event

_event_listeners_enabled = False


def enable_event_listeners():
    global _event_listeners_enabled

    with threading.Lock():
        if not _event_listeners_enabled:
            _enable_sqlalchemy_event_listeners()
            _enable_alembic_event_listeners()
            _event_listeners_enabled = True


def _enable_sqlalchemy_event_listeners():
    from sqlalchemy_postgresql_audit.event_listeners.sqlalchemy import (
        create_audit_table,
    )

    event.listens_for(Table, "after_parent_attach")(create_audit_table)


def _enable_alembic_event_listeners():
    try:
        from sqlalchemy_postgresql_audit.event_listeners.alembic import (
            compare_for_table,
        )
        from alembic.autogenerate.compare import comparators

        comparators.dispatch_for("table")(compare_for_table)
    except ImportError:
        pass



"""Defines event listeners for:

    - creating table objects
    - generating trigger/procedure DDL for audit tables.

"""
from sqlalchemy import Column, DateTime, String, Table

from sqlalchemy_postgresql_audit.ddl import (
    get_audit_spec,
    get_create_trigger_ddl,
    get_drop_trigger_ddl,
)
from sqlalchemy_postgresql_audit.dialect import (
    DEFAULT_AUDIT_TABLE_FUNCTION_NAMING_CONVENTION,
    DEFAULT_AUDIT_TABLE_NAMING_CONVENTION,
    DEFAULT_AUDIT_TABLE_TRIGGER_CONVENTION,
)


def create_audit_table(target, parent):
    """Create an audit table and generate procedure/trigger DDL.

    Naming conventions can be defined for a few of the named elements:

        - audit.table: Controls the name of the table
        - audit.function: Controls the name of the function
        - audit.trigger: Controls the name of the trigger on the table

    This function creates a new companion table to store row versions.

    Any :class:`sqlalchemy.sql.schema.Column`s specified in `table.info['session_settings']` will
    be copied and included in the audit table.

    This function will leave a key in the audited table:

        .. code-block:: python

            table.info['audit.is_audited']

    And a key in the audit table:

        .. code-block:: python

            table.info['audit.is_audit_table']

    Additionally you can find the relevant create/drop ddl at the followng keys:

        .. code-block:: python

            table.info['audit.create_ddl']
            table.info['audit.drop_ddl']

    :param target: The :class:`sqlalchemy.sql.schema.Table` to make an audit table for
    :param parent: The :class:`sqlalchemy.sql.schema.MetaData` to associate the audit table with.
    :return: None
    """
    audit_spec = get_audit_spec(target)

    if not audit_spec.get("enabled"):
        return

    audit_table_naming_convention = parent.naming_convention.get(
        "audit.table", DEFAULT_AUDIT_TABLE_NAMING_CONVENTION
    )
    audit_function_naming_convention = parent.naming_convention.get(
        "audit.function", DEFAULT_AUDIT_TABLE_FUNCTION_NAMING_CONVENTION
    )
    audit_trigger_naming_convention = parent.naming_convention.get(
        "audit.trigger", DEFAULT_AUDIT_TABLE_TRIGGER_CONVENTION
    )

    audit_table_name = audit_table_naming_convention % {
        "table_name": target.name,
        "schema": audit_spec["schema"] or "public",
    }
    audit_function_name = audit_function_naming_convention % {
        "table_name": target.name,
        "schema": audit_spec["schema"] or "public",
    }
    audit_trigger_name = audit_trigger_naming_convention % {
        "table_name": target.name,
        "schema": audit_spec["schema"] or "public",
    }

    columns = [
        Column(col.name, col.type, nullable=True) for col in target.columns.values()
    ]
    session_setting_columns = [col.copy() for col in audit_spec["session_settings"]]
    for col in session_setting_columns:
        col.name = "audit_{}".format(col.name)

    column_elements = session_setting_columns + columns

    audit_table = Table(
        audit_table_name,
        target.metadata,
        Column("audit_operation", String(1), nullable=False),
        Column("audit_operation_timestamp", DateTime, nullable=False),
        Column("audit_current_user", String(64), nullable=False),
        *column_elements,
        schema=audit_spec["schema"]
    )

    target.info["audit.create_ddl"] = get_create_trigger_ddl(
        target.columns,
        audit_table.columns,
        audit_function_name,
        audit_trigger_name,
        target.fullname,
        audit_table.fullname,
        session_setting_columns,
    )

    target.info["audit.drop_ddl"] = get_drop_trigger_ddl(
        audit_function_name, audit_trigger_name, target.fullname
    )
    audit_table.info["audit.target_table"] = target

    audit_table.info["audit.is_audit_table"] = True
    target.info["audit.is_audited"] = True



from sqlalchemy.dialects.postgresql.base import RESERVED_WORDS

from .templates import make_audit_procedure, make_drop_audit_procedure


def get_audit_spec(table):
    audit_spec = table.info.get("audit.options", {"enabled": False})

    audit_spec["schema"] = audit_spec.get("schema_name", table.schema)
    audit_spec["session_settings"] = audit_spec.get("session_settings", [])

    return audit_spec


def get_create_trigger_ddl(
    target_columns,
    audit_columns,
    function_name,
    trigger_name,
    table_full_name,
    audit_table_full_name,
    session_settings=None,
):
    session_settings = session_settings or []

    deletion_elements = ["'D'", "now()", "current_user"]

    updation_elements = ["'U'", "now()", "current_user"]

    insertion_elements = ["'I'", "now()", "current_user"]

    setting_map = {
        session_setting.name: session_setting for session_setting in session_settings
    }

    column_elements = []
    check_settings = []

    for col in audit_columns.values():
        column_name = (
            '"{}"'.format(col.name)
            if (col.name.lower() in RESERVED_WORDS or col.name.lower() != col.name)
            else col.name
        )

        # We need to make sure to explicitly reference all elements in the procedure
        column_elements.append(column_name)

        # If this value is coming out of the target, then we want to explicitly reference the value
        if col.name in target_columns:
            deletion_elements.append("OLD.{}".format(column_name))
            updation_elements.append("NEW.{}".format(column_name))
            insertion_elements.append("NEW.{}".format(column_name))

        # If it is not, it is either a default "audit_*" column
        # or it is one of our session settings values
        else:
            if col.name in (
                "audit_operation",
                "audit_operation_timestamp",
                "audit_current_user",
            ):
                continue

            session_setting = setting_map[col.name]
            type_str = session_setting.type.compile()
            name = session_setting.name.split("audit_", 1)[-1]

            session_settings_element = "current_setting('audit.{}', {})::{}".format(
                name, "true" if session_setting.nullable else "false", type_str
            )
            deletion_elements.append(session_settings_element)
            updation_elements.append(session_settings_element)
            insertion_elements.append(session_settings_element)

            # This handles a kind of strange behavior where if you set a session setting
            # and then commit the transaction you will end up with an empty string in that setting
            # and then the procedure will succeed, despite the value being "empty".
            if not session_setting.nullable:
                check_settings.append(
                    "IF {}::VARCHAR = '' THEN RAISE EXCEPTION "
                    "'audit.{} session setting must be set to a non null/empty value'; "
                    "END IF;".format(session_settings_element, name)
                )

    return make_audit_procedure(
        audit_table_full_name=audit_table_full_name,
        table_full_name=table_full_name,
        procedure_name=function_name,
        trigger_name=trigger_name,
        deletion_elements=deletion_elements,
        updation_elements=updation_elements,
        insertion_elements=insertion_elements,
        audit_columns=column_elements,
        check_settings=check_settings,
    )


def get_drop_trigger_ddl(function_name, trigger_name, table_full_name):
    return make_drop_audit_procedure(function_name, trigger_name, table_full_name)


def install_audit_triggers(metadata, engine=None):
    """Installs all audit triggers.

    This can be used after calling `metadata.create_all()` to create
    all the procedures and triggers.

    :param metadata: A :class:`sqlalchemy.sql.schema.MetaData`
    :param engine: A :class:`sqlalchemy.engine.Engine` or None
    :return: None or a :class:`str` for the DDL needed to install all audit triggers.
    """
    audit_table_ddl = [
        t.info["audit.create_ddl"]
        for t in metadata.tables.values()
        if t.info.get("audit.is_audited")
    ]

    engine = engine or metadata.bind

    if engine:
        for ddl in audit_table_ddl:
            engine.execute(ddl)
    else:
        return "; ".join(audit_table_ddl)


def uninstall_audit_triggers(metadata, engine=None):
    """Uninstalls all audit triggers.

    This can be used to remove all audit triggers.

    :param metadata: A :class:`sqlalchemy.sql.schema.MetaData`
    :param engine: A :class:`sqlalchemy.engine.Engine` or None
    :return: None or a :class:`str` for the DDL needed to uninstall all audit triggers.
    """
    audit_table_ddl = [
        t.info["audit.drop_ddl"]
        for t in metadata.tables.values()
        if t.info.get("audit.is_audited")
    ]

    engine = engine or metadata.bind

    if engine:
        for ddl in audit_table_ddl:
            engine.execute(ddl)
    else:
        return ";\n ".join(audit_table_ddl)
